#include "tablet.h"
#include <stdio.h>
#include <string.h>

status_t
convertpositiontostruct(const uint8 *data, tablet_position &);
status_t
convertinfotostruct(const uint8 *data, tablet_info &);
static long 			__reader(void *arg);
static bool	reader_quit = false;
static bool	reader_ready = false;
thread_id	reader_id;

const uint8 kSyncBit		= 7;
const uint8 kProximityBit 	= 6;
const uint8 kPointerBit		= 5;
const uint8 kButtonBit		= 3;
const uint8 kPressureSignBit	= 6;

const uint8	kSyncBitMask = 0x1 << kSyncBit;
const uint8 kProximityBitMask = 0x1 << kProximityBit;
const uint8 kPointerBitMask = 0x1 << kPointerBit;
const uint8 kButtonBitMask = 0x1 << kButtonBit;
const uint8 kPressureSignBitMask = 0x1 << kPressureSignBit;

status_t
convertinfotostruct(const uint8 *data, tablet_info &info)
{	
}

status_t
convertpositiontostruct(const uint8 *data, tablet_position &info)
{	
	if (!(data[0] & kSyncBitMask))
	{
		printf("convertpositiontostruct - early return, no sync bit\n");
		return B_ERROR;
	}
	
	info.proximity = (data[0] & kProximityBitMask) >> kProximityBit;
	info.stylus = (data[0] & kPointerBitMask) >> kPointerBit;
	info.buttonflag = (data[0] & kButtonBitMask) >> kButtonBit;

	info.x = (data[0] & 0x3) << 14;
	info.x += (data[1] & 0x7f) << 7;
	info.x += (data[2] & 0x7f);
	
	info.buttons = (data[3] & 0x78) >> 3;
	
	info.y = (data[3] & 0x3) << 14;
	info.y += (data[4] & 0x7f) << 7;
	info.y += (data[5] & 0x7f);
		
	info.pressuresign = (data[6] & kPressureSignBitMask) >> kPressureSignBit;
	info.pressuredata = (data[6] & 0x3f);
}

BTSWacomTablet::BTSWacomTablet(const char *port)
{
	fPort.SetDataRate(B_9600_BPS);
	fPort.SetDataBits(B_DATA_BITS_8);
	fPort.SetStopBits(B_STOP_BITS_1);
	fPort.SetParityMode(B_NO_PARITY);
	fPort.SetFlowControl(B_NOFLOW_CONTROL);
	fPort.SetBlocking(false);
	
	fPortStatus = fPort.Open(port);
	fPort.SetTimeout(250000);		// Timeout of quarter a second
	//printf("Open Status: %d\n", fPortStatus);

	SelectIVCommandSet();
	
	// Start port reading thread
	//resume_thread(reader_id = spawn_thread(__reader,"wacom_reader",B_NORMAL_PRIORITY,&fPort));	
	reader_ready = true;
	// Don't Continue until thread is ready
	// Block on semaphore waiting to be released.
	while(!reader_ready)
		;
		
	strcpy(fCommandTerminator, "\r\n");
}

BTSWacomTablet::~BTSWacomTablet()
{
	int32 dummy;
	
	reader_quit= true;
	fPort.Close();
	//wait_for_thread(reader_id,&dummy);
}

void	
BTSWacomTablet::SendCommand(const char *cmd)
{
	fPort.Write(cmd,strlen(cmd));
}

//
// Retrieve a line of data from the tablet.  It should be
// terminated with a <cr>
void
BTSWacomTablet::GetLine(char *buff, const long buffLen)
{
	//return;
	
	bool done = false;
	char *ptr = buff;
	uint8 aChar;
	long bytestoread;
	long totalread = 0;
	
	while(!done && (totalread < buffLen))
	{
		bytestoread = fPort.WaitForInput();
		//printf("BytesToRead - %d\n", bytestoread);
		while(bytestoread)
		{
			long numRead = fPort.Read(&aChar, 1);
			if (numRead>0)
			{
				//printf("%c",aChar);
				//fflush(stdout);
				if (aChar == '\r')
				{
					*ptr = '\0';
					done = true;
					break;
				} else
					*ptr = aChar;
				ptr++;
				totalread ++;
			}	
			bytestoread -= numRead;
		}
	}	
}

status_t
BTSWacomTablet::GetBytes(uint8 *bytes, const uint32 numBytes)
{
	long bytestoread = numBytes;
	long numRead;
	uint8 *buffPtr = bytes;
	
	long byteswaiting = fPort.WaitForInput();
	if (byteswaiting < 1)
	{
		return B_ERROR;
	}

	while (bytestoread)
	{
		numRead = fPort.Read(buffPtr, bytestoread);
		buffPtr += numRead;
		bytestoread  -= numRead;
	}

	return B_NO_ERROR;
}


status_t
BTSWacomTablet::GetModel(char *buff)
{
	buff[0] = '\0';
	
	SendCommand("~#\r");
	GetLine(buff,32);

	return B_NO_ERROR;
}

status_t	
BTSWacomTablet::GetMaxCoordinates(int32 &x,int32 &y)
{
	char buff[32];
	
	SendCommand("~C\r");
	GetLine(buff,32);
	//printf("Max Coordinates: %s\n", buff);
	
	return B_NO_ERROR;
}

status_t
BTSWacomTablet::GetCurrentSettings(char *buff, const int32 buffLen)
{
	SendCommand("~R\r");
	GetLine(buff,buffLen);
	
	return B_NO_ERROR;
}

status_t	
BTSWacomTablet::GetTabletInfo(tablet_info &info)
{
	// Send the individual request command
	SendCommand("RQ1\r");
	
	// Try to get the position information
	uint8 buff[14];
	status_t status = GetBytes(buff, 7);
	
	if (status == B_ERROR)
	{
		//printf("No bytes read.\n");
		return status;
	}

	//status = convertinfotostruct(buff, info);
	
	if (status == B_ERROR)
	{
		printf("Invalid Data.\n");
		return status;
	}
		
	return B_NO_ERROR;
}

status_t	
BTSWacomTablet::GetPosition(int &x, int &y, int &buttons)
{	
	tablet_position info;
	GetPosition(info);
	
	x = info.x;
	y = info.y;
	buttons = info.buttons;
	
	return B_NO_ERROR;
}

status_t	
BTSWacomTablet::GetPosition(tablet_position &position)
{	
	uint8 buff[128];

	SendCommand("RQ1\r");
	long numRead = fPort.Read(buff, 7);
	convertpositiontostruct(buff, position);
	
	return B_NO_ERROR;
}

status_t
BTSWacomTablet::SetASCIIFormat()
{
	SendCommand("AS0\r");
}

status_t
BTSWacomTablet::SetBinaryFormat()
{
	SendCommand("AS0\r");
	return B_NO_ERROR;
}

status_t
BTSWacomTablet::SetPointMode()
{
	SendCommand("PO\r");
	return B_NO_ERROR;
}

status_t
BTSWacomTablet::SetRelativeMode()
{
	SendCommand("DE1\r");
	return B_NO_ERROR;
}

status_t
BTSWacomTablet::SetAbsoluteMode()
{
	SendCommand("DE0\r");
	return B_NO_ERROR;
}

status_t
BTSWacomTablet::SelectIISCommandSet()
{
	SendCommand("$\r");
	return B_NO_ERROR;
}

status_t
BTSWacomTablet::SelectIVCommandSet()
{
	SendCommand("#\r");
	return B_NO_ERROR;
}

status_t
BTSWacomTablet::Select1201CommandSet()
{
	SendCommand("@\r");
	return B_NO_ERROR;
}

status_t
BTSWacomTablet::SetAlwaysTransmit(const int8)
{
	SendCommand("AL%d\r");
	return B_NO_ERROR;
}

status_t	
BTSWacomTablet::SetTransmissionInterval(const int16)
{
	return B_NO_ERROR;
}


status_t
BTSWacomTablet::Reset()
{
	SendCommand("RE\r");
	fPort.ClearInput();
	fPort.ClearOutput();
	return B_NO_ERROR;
}

// We need a separate reading thread which will take care of just
// sitting around waiting for input.  This way, when there is input
// it can be sent immediately to the receiver.  In some cases it 
// will be the response to a command, and in others it will be
// the continuous stream of positional information.

//=======================================================
// Independent reading thread
//=======================================================
long __reader(void *arg)
{
	BSerialPort    *port = (BSerialPort*)arg;
	
	//printf("__reader\n");
	char buff[10];
	int numread=0;
	long bytestoread; 
	
	reader_ready = true;
	while (!reader_quit) 
	{
		bytestoread = port->WaitForInput();
		while(bytestoread)
		{
			numread = port->Read(buff,1);
			if (numread < 1)
			{	
				printf("__reader - numread < 1\n");
				break;
			}	
			printf("%c",buff[0]);	
			bytestoread -= numread;
		}
		fflush(stdout);
	}

	return 0;
}


